/*-----------------------------------------------------------------------
 * Copyright 1999 Michael Lincoln <mclinc@essex.ac.uk>
 * 
 * This software is provided free of charge.  Permission is hereby given
 * to use, copy, modify and distribute free of charge the software for
 * any purpose provided that this notice is included verbatim in every
 * copy.  The software is provided on an "as is" basis and the authors
 * and the University of Essex have no obligation to provide maintenance,
 * support, updates, enhancements, or modifications.
 * 
 * All express or implied warranties whether statutory or otherwise as to
 * the merchantability or fitness for a particular purpose of the
 * software are, to the maximum extent permitted by law, excluded.
 *
 * In no circumstances whatsoever shall the author or the University of
 * Essex be liable in contract, tort, or otherwise including liability
 * for negligence for any loss whether direct, indirect or otherwise of
 * business, revenue, profits, anticipated savings, wasted expenditure,
 * corruption or destruction of data or for any indirect or consequential
 * loss whatsoever.
 * 
 *
 *------------------------------------------------------------------------
 */





/*
  VISCA class provides API framework for using the VISCA communications 
  protocol
  Multipul devices (upto 7) may be connected to one physical bus each physical
  bus has one visca object associated with it.

  VISCA devices may receive upto two commands in there buffers.
  
  Support for a single cam is now fully implimented using both sockets
  multiply cam support should be easy to add by extending the socketsInuse
  array and adding a bit of code to unintertingPacket method.
  But I dont have a secound cam to test this with , so I haven't bothered.

  V0.01  extracted from my MSc project for use else where 10 sept 98

  v0.01  found bug in the speed setting for the moves which was in decimal
         when it should have been hex, leading to slower than nessasy moves

  v0.02  Rewriten to accept comport name in the constructor
  



*/

#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <termios.h>
#include <stdlib.h>
#include <iostream.h>
//#include "cam.h"

//setup the communications device exit on fault, as this is fatal
visca::visca(char *comname)
{
  myaddress =0; //the address of the controling device (me). I havent made this 
  // changerble because I cant see having two computers controling one cammera.
  // but if's its needed you can always add another method.
  //but dont expect the rest of the code to work correctly if there are two controlers
  
  // use /dev/cua? for linux
  //    /dev/cuaa? for FreeBSD
  //    /dev/ttyd? for irix

  
  //const char* comname ="/dev/cuaa1";
  
  //open the serial device for read write

  
  comdev = open(comname ,O_RDWR|O_NOCTTY);//|O_NONBLOCK);

  
  if(comdev<0)
    {
      cerr <<"Visca: Error, opening comport :"<<comname<<endl;
      exit(1);
    }
  
  if(isatty(comdev)==0)
    {
      cerr <<"Visca: Error, "<<comname<<" is not a tty!"<<endl;
      exit(1);
    }
  
  //read the current terminal settings and save them so that they can be restored
  //on exit (in the distructor)
  tcgetattr(comdev,&savedSettings);
  
  //now get another copy to modify to the new settings we need.
  tcgetattr(comdev,&newSettings);
  
  //set the terminal to operate in the BSD-like raw mode (ie, no nasty traslation of lf cr 
  //esc's etc.) this also sets it to 8 data bits, no parity, 1 stop bit. 
  // cfmakeraw(&newSettings);
  
  //set the baud rate to 9600 in a posix friendly way.
  cfsetospeed(&newSettings,B9600);
  cfsetispeed(&newSettings,B9600);
  
  
  //parranoid settings
  
  

  //things to turn on ignore brake.
  newSettings.c_iflag |= IGNBRK;
  //things to turn all other striping, translation, etc.
  newSettings.c_iflag &= ~(INPCK|ISTRIP|IGNCR|ICRNL|INLCR|IXOFF|IXON|IMAXBEL);
  
  
  //configure output modes
  //turn off all output translation (for BSD systems only)
  newSettings.c_oflag &= ~(OPOST|ONLCR);
  
  
  //set local modes for terminal
  //Turn off echoing and some signal handeling
  newSettings.c_lflag &= ~(ICANON|ECHO|ISIG|IEXTEN);
  
  
  
  //set the time out before returning 
  newSettings.c_cc[VMIN]  = 0; //number of bytes we need to return (0 means just 1 will do)
  newSettings.c_cc[VTIME] = 1; //time out in 0.1 secounds if no data received within this time
                         //a read will return with -1 and errno == EAGAIN to indicate no data
  
  //VTIME is set to 100ms so that the number of timeouts can be counted allowing 
  //different timeouts to be given to waiting for the start of a packet and the time waited
  //between bytes.  Without the need to keep resetting the terminal device (and then waiting
  //for it to drane or flush)
  
  //now make the changes, flushing away any data already in the buffers
  
  if(tcsetattr(comdev,TCSAFLUSH,&newSettings)<0)
    {
      cerr<<"Visca: Can't set terminal info"<<endl;
      exit(1);
    }
  
  socketInuse[1] = false; //assume that the cammera is idle when the program is started
  socketInuse[2] = false; //a bit cheep, but it'll work
  
}


visca::~visca()
{
  //restore the settings as this class found them (the shell _should_ do this).
  tcsetattr(comdev,TCSAFLUSH,&savedSettings);

  //close the terminal device
  close(comdev);
} 


/*
  send a packet over the serial line. the unsigned char array MUST be at least 
  data_length bytes long.  The address to send it to should be specifyed with 
  you guessed it address, which should be between 0 and 7.The type should be either
  COM or INQ (which are constant public members of visca)
  
  only returns _after_ the data has left the terminals buffer.  It returns true
  if successfull and false if there was a problem. This could be a hardware problem
  or more likely an invalid parameter.
*/

int visca::send_packet(int address,int type,int data_length,unsigned char* data)
{
  
  // cerr <<" visca,send_packet called"<<endl;
  
  unsigned char packet[16];
  
  //set the from address to 1 and the too addr to address
  packet[0] = 0x80 |((unsigned char)address&0x0F);
  packet[1] = 0x0F & (unsigned char) type; //mask off type just to be parranoid
  
  packet[2+data_length] = 0xFF ; //add the terminator
  
  //copy the data into the packet
  for(int n=2;n<data_length+2;n++)
    packet[n]= *(data++);  
  
  if(write(comdev,(void*) packet,data_length+3)>=0) //send the packet
    {
      tcdrain(comdev); //wait for the transmit buffer to empty.
      return true; //return, we did it!
    }
  
  return false; //failed to transmitt bytes.
}


/*
  receives a packet from a cam on the line.  Waits for a preset time out before giving up.
  unsigned char array should be 16 bytes longs to allow for the maximum size packet which 
  can be sent over a visca network.  data_length will be set to the length of the data
  received. This will be zero if it was an ack or compleation packet.
  
  returns true if a vaild packet was received, otherwise returns false
  
*/
int visca::recv_packet(int &address,int &type,int &socket,int &data_length,unsigned char* data)
{
  // cerr <<"visca, recv_packet called"<<endl;
  unsigned char packet[17]; //space for the packet
  unsigned char abyte;
  int bytesReceived = 0; //number of bytes received not including the FF terminator
  
  int timer = 0;
  int debug  = 0;
  
  //  received the packet give up if timeout and finish when terminator found
  
  while(bytesReceived<17)
    {
      //read will return after 0.1 secound if no data is avalable
      while(read(comdev,&abyte,1)!=1) //keep trying too read a byte, until we do.
	if(timer++>1) 
	  {
	    // cerr<<"visca,Timeout,this may be normal! debug flag="<<debug<<endl;
	    return false; //give up after .30 secound (or 3 retrys) 10ms is too small
	  }
      debug = 1;
      timer = 1; //now only wait 0.2 of a secound between bytes in a packet this should 
      //NOT timeout too easly, it causes LOT of syncronisation problems
      //and as i have no error recovery the program must exit.
      
      packet[bytesReceived++] = abyte;
      
      if(abyte == 0xFF) break; //exit while loop if its a terminator 
    }
  
  if(bytesReceived==17)
    {
      cerr<<"visca, Over sized packet received"<<endl;
      return false; // packets cannot legaly be this big
    }
  
  /*
    Now dismanel the packet as far as possible (data link layer if you like osi 7 layer
    model 9-) (I think its a wast of time, the model has to be bent to find every practical 
    protocol, but then I'm just a jumped up pleb 8-)
  */
  
  if (bytesReceived>=3)  //if its large enough to be a valid packet
    {
      if((packet[0] & 0x0F) == myaddress) //if its sent to me
	{
	  address = (0x0F & ( packet[0]>>4)) - 8; //extract the upper nible of the 1st byte and - 8
	  
	  //the type is the upper nibble of the secound byte
	  type = (0x0F & (packet[1]>>4));
	  
	  //the lower nibble of that byte is the socket number (see visca docs)
	  socket = 0x0F & packet[1];
	  
	  // bytesReceived - 3 is the number of bytes of data (normaly zero )
	  data_length = bytesReceived - 3;
	  
	  //now do some syntax checking to make shore noting nosencical is being done
	  
	  //data length will be zero if its an ack or command compleation
	  //the only exception is if its a responce to an inquiry  or error message.
	  //in which case it must be of type CMP or ERR (0x5X/0x6X on byte 1),
	  //lets just test for this just in case
	  
	  //still a bug below , when error message received, it causes false syntax error!!!
	  // if((data_length>0 ) && ((type != CMP )||(type != ERR)))
	  // {
	  //   cerr<<"Syntax error, data length too big for type"<<endl;
	  //   return false; //oh dear, seems somethings gone wrong, panic.
	  // }
	  
	  if((type == ERR) &&(data_length!=1))
	    {
	      cerr << "visca, packet syntax error, data length must be 1 if error message rx'ed" <<endl;
	      return false;
	    }
	  
	  
	  //else carry on and copy accross the data.

	  for(int n=2;n<bytesReceived-1;n++) //copy from byte 2 to last but 1 (the terminator)  inclusive
	    {

	      *(data++) =packet[n];
	    }
	  
	  //success
	  return true;
 
	}
    }
  cerr <<"visca, packet syntax error, not addressed to me or too small"<<endl;
  return false;
}


//send a command and sort out the acks and compleations then get returned
//this is the next layer up in the protocol stack

int visca::send_command(int address,int data_length,unsigned char *data)
{
  //started at 1:50am 1/4/97
  
  //this rather complex procedure sends a command and waits for an
  //acknolagement back.  It returns before receiving the compleation
  //packet from the cam. (this is delt with when this method is next called)
  
  //the reasion for the convoluted method used, is that both sockets in the 
  //cammera are used, therefor allowing two commands to be exicuted at the same
  //time. ie, pan&tilt + zoom, which speeds things up alot.

  //see the flow chart in the documentation for an explination

  //!NB: the address from which the packets is sent from is not currently tested.
  //     it is assumed you only have one cam connected to the visca network
  //     support for addional cammeras can given by adding another dimention 
  //     to the socketInuse array to allow seperate socket statuses to be maintained
  //     and tested for. This is quite easy, but I'm walking before running, and I cant
  //     test it with two cams I dont have two 8-(

  // I have NOT implimented the command cancel packet type, this provents me having
  // to deal with canceled commands.  Commands will allways be compleated, barring 
  // acts of God. This saves a LOT of code, and makes it lots nicer to read.
  // If you dont want a command to be carried out, dont send it 9-)
  // If you realy want to stop a command , then it could be implimented in a send_stop
  // method, which would then deal with the command canelled error messages etc. as 
  // long as it prommised this method would never see one, I would be happy.

  
  //are there any packets in the comports buffer ??
  
  int rxedAddress,rxedType,rxedSocket,rxedDataLength;
  unsigned char rxedData[17];
  
  //cerr <<"send_command started, dealing with packets in buffer"<<endl;

  //while we have valid packets to decode , do something with them
  //if not then continue on our way and send the command, unless there are no
  //free sockets inwhich case, just keep waiting untill one becomes free
  do
    {
      //cerr <<"about to see if anything in buffer"<<endl;
      //we would like to sort out any outstanding packets before we start sending more
      //commands
      while(recv_packet(rxedAddress, rxedType,rxedSocket,rxedDataLength,rxedData))
	{
	  //were not interested in these packets
	  if(!uninterestingPacket(rxedAddress,rxedType,rxedSocket,rxedDataLength,data)) 
	    return false; //if it cant deal with it give up.
	} 
      
      //now we've finished dealing with all the packets that where qued up
      
    }  while (socketInuse[1]&&socketInuse[2]); //but keep waiting if we have no free sockets
  
 
  // we have at least one free socket, so continue.
  
  // cerr <<"Finished dealing with packets in buffer, sending command"<<endl;
  
  //lets send the command
  
  
  if(!send_packet(address,COM,data_length,data))
    { 
      //either send_packet method barfed at one of its parramiters or the coms port
      //has exploded 
      cerr <<"visca, Could not send command packet"<<endl;
      exit(1);
    }
  
  //the acknolagement should be sent back at onces
  //there is a (very) small change that an unconnected  packet could have been received
  //between the time that the buffer was last checked and now, so check just in case.
  
 // cerr <<"visca, command sent waiting for ack"<<endl;
  
  //receive the packet
  do
    {
      if(!recv_packet(rxedAddress,rxedType,rxedSocket, rxedDataLength,rxedData))
	{
	  cerr << "visca,No acknolagement received from cam"<<endl;
	  return false; //we did not get an ack back
	}
      
      //cerr <<"packet receved was of type " <<rxedType<<endl;
      
      //have a look to see if its an acknolagement
      
      if(rxedType==ACK && (1<=rxedSocket<=2) )
	{
	  //if so, mark the socket as in use.
	  socketInuse[rxedSocket]=true;
	  //and return happy
	  return true;
	  
	}
      else  
	{
	  //its a packet we didn't realy want but was sent asyncronusly to use. Joy
	  //this has just got a lot easyer to deal with now I've added the uninterestedPacket 
	  //class
	  if(!uninterestingPacket(rxedAddress,rxedType,rxedSocket,rxedDataLength,data))
	    //if it couldn't sorted it out return with bad news
	    return false;
	}
    } 
  while(1); //try again until something happens
  
  
  return false; //this will never happed just fancyed wasting some bytes of core...
  
}




//send an inquirire to the cam ,and return the data that was sent back

int visca::send_inquirie(int address,unsigned char *inquirie,int &rxedDataLength,unsigned char *data)
{
  
  //send out the inquirie
  if(!send_packet(address,INQ,2,inquirie)) 
    {
      cerr<<"visca, cant send inquirie"<<endl;
      return false;
    }
  
  int rxedAddress,rxedType,rxedSocket;
  
  do
    {
      if(!recv_packet(rxedAddress,rxedType,rxedSocket,rxedDataLength,data))
	{
	  cerr<< "no reply to inquirie"<<endl;
	  return false;
	}
      else 
	{
	  //make shore its a reply to an iquirie, if not pass it to uninterestingPacket, so it
	  //can deal with it
	  if((rxedType!= CMP)||(rxedDataLength<1))
	    {
	      if(!uninterestingPacket(rxedAddress,rxedType,rxedSocket,rxedDataLength,data))
		{
		  //if that can't deal with it , then we've had it,
		  cerr <<"visca, unexpected reply to inquirie,cannot be delt with!"<<endl;
		  return false;
		}
	    }
	  else 
	    return true;  //we did it
	}
    }
  while(true);  //keep trying unill we get a packet or give up.
  
  return false; //never happen
  
}

//deals with returened packets that wern't wanted (sob) and gives them a home..
//it updates the status of the sockets when they become free and sends out network reset packets
//we a new device is added to the bus (a bit pointless because the other methods cant deal with 
//mulitple devices yet, but they are very close to being able.

int visca::uninterestingPacket(int rxedAddress,int rxedType,int rxedSocket,int rxedDataLength,unsigned char *rxedData)
{

  
  switch( rxedType)
    {
    case CMP: //if it was a compleation packet then mark that socket as free
      {
	if(rxedSocket == 0)
	  {
	    //compleation packets from socket 0 are replys to inquirys the only
	    //method which should cause these should be  send_inqurie  and this would
	    //be Very interested in it(!) so this must be an error, 
	    cerr <<"visca, inquiry reply  deamed uninteresting, this shouldn't happen" <<endl;
	    return false;
	  }
	
	
	if(1<=rxedSocket<=2) 
	  {
	    //if its a compleation packet for a valid socket, then free that socket so
	    //that packets may be sent to it.
	    socketInuse[rxedSocket]=false;
	    return true; //this is good, return
	  }
	else 
	  {
	    //this shouldnt happen
	    cerr<<"visca, invalid socket number received"<<endl;
	    return false;
	  }
	break;
      }
      
    case ERR: 
      
      {
	//display the error number and return with a fault, not bothering with any error
	//recovery tatics at the moment, they never seem to work anyway.
	
	cerr <<"visca, Error message received from device,method uninterested, error code > "
	     << (int) rxedData[0]<<endl;
	return false;
	break;
      }
      
    case NET: //if its a network change, you must send a network reset packet
      {
	//give a warning we cant realy deal with this is untested
	cerr <<"visca,NB:network has changed, may not notice new device!"
	     <<endl;
	//broadcast the network reset packet, this will cause the devices
	//to number themselfs with addresses. (its not the switch on the back
	//of the cam that sets the address!! (nice of them you tell you in the
	//manual 9-))
	unsigned char message = (unsigned char) 0x01; //the magic message 
	if(!send_packet(8,RST,1,&message)) exit(1); //sending shouldnt fail, 
	
	return true; //delt with return happy
	
	break;
      }
      
    case ACK: //you shouldn't ever get an ack packet here 
      {
	//ack packets are only generated after sending commands, in the send command
	//method. that method will be interested in them, so they shouldn't get here
	
	cerr <<"visca, an acknolagement has been deamed uninteresting, this shouldn't happen"
	     <<endl;
	return false;
      }
      
      
    default:  //an unrecogised packet type has been received
      {
	//either sony have added somethings without telling me, or line noise
	
	cerr <<"visca, Unrecogised packet type received"<<endl;
	return false;
      }
    }
  
  //this will NEVER happen

  return false;
  
}












/*
The cam class, the next object up in the stack. provides some usefull high level
 functions to access the cam.
To dos...
*make it take notice of the cam address.the visca object can use diffent addresses but can 
only talk to one cam.  if only one cam on the bus, then its address will be 1...

*add more funtions, including some method of waiting for a zoom to compleat, there is
no was of asking the cam this, but you can inquire about its current zoom setting, which 
could be used.

*add a retry so that failed commands are resent a few times, this seems to work 90%
 of the time even though the visca object just detects errors, it dosnt try to fix 
them.(it will do a network reset when it detects some form of transmition fault 
(abit like scsi) when i get round to adding it.)


 **now accepts com port name in the constructor


*/

cam::cam(char* portname,int camAddress)
{
 
  myCamAddress = camAddress;

  myViscaPTR = new visca(portname);
  if(!myViscaPTR)
    {
      cerr << "cam:Memory allication fault"<<endl;
      exit(1);
    }

  lastX = lastY = lastZ = 0;
  
  
}

cam::~cam()
{
 if(myViscaPTR) free(myViscaPTR);
  
}



//NB, this can only turn it on if its pluged in 9-)

int cam::powerOn(int turnOn)
{
  unsigned char command[]={0x04,0x00,0x03}; // command to turn off cam

  //if we want to turn it on, modify the command
  if(turnOn) command[2] =(unsigned char) 0x02; 


  //now send it 
  int status = myViscaPTR->send_command(1,3,command);
  
  return status;
}


int cam::move(int x,int y)
{
  
  //add test to to make shore that the x and y values are valid

  //mask z to 16 bits (this only realy makes a difference when compairing the
  //last possition to the new onein waitCompleate.

  x &= 0xFFFF;
  y &= 0xFFFF;

  //define the command to make the cam move
  unsigned char command[12] = {0x06,0x02,0x18,0x14,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00};

  //now fill in the bytes for the point to move too
  command[4] = (unsigned char) (0x0f & (x>>12));
  command[5] = (unsigned char) (0x0f & (x>>8));
  command[6] = (unsigned char) (0x0f & (x>>4));
  command[7] = (unsigned char) (0x0f & x);
	
  command[8]  = (unsigned char) (0x0f & (y>>12));
  command[9]  = (unsigned char) (0x0f & (y>>8));
  command[10] = (unsigned char) (0x0f & (y>>4));
  command[11] = (unsigned char) (0x0f & y);
  
  //now send it 
  int status = myViscaPTR->send_command(1,12,command);

  //remember where it should end up if it worked.
  if(status) 
    {
      lastX = x;
      lastY = y;
    }



  return status;
}


int cam::zoom(int z)
{
  //define the zoom command, with zeros for the data
  unsigned char  command[]={0x04,0x47,0x00,0x00,0x00,0x00};

  

  //fill in the data with the z value
  command[2] = ( unsigned char) (0x0f & (z>>12));
  command[3] = ( unsigned char) (0x0f & (z>>8));
  command[4] = ( unsigned char) (0x0f & (z>>4));
  command[5] = ( unsigned char) (0x0f & z);
	
  //send it
  
  int status = myViscaPTR->send_command(1,6,command);
 
  //remember where it should end up if it worked.
  if(status) lastZ = z;

  return status;
}



//ask the cam where its pointing

int cam::getPos(int &x,int &y)
{
  unsigned char Inquirie[]={0x06,0x12};//Pan-titlPosInq 
  
  int rxedDatalen;
  unsigned char Data[17];
  
  //Inquirie to the position of the pan & tilt.
  //give up if it fails
  if(!myViscaPTR->send_inquirie(1,Inquirie,rxedDatalen,Data)) return false; 
    if(rxedDatalen!=8) return false; //must be 8 bytes
  //the format of the data is.. 0x 0x 0x 0x 0y 0y 0y 0y

  //extract the x pos from the received data
  x = (0x0f&Data[3]) | ((0x0f&Data[2])<<4) | ((0x0f&Data[1])<<8) | ((0x0f&Data[0])<<12);
  y = (0x0f&Data[7]) | ((0x0f&Data[6])<<4) | ((0x0f&Data[5])<<8) | ((0x0f&Data[4])<<12);
  
  //make it signed
  //if (x>0x370) x=(0xffff -x)*-1;
  //if (y>0x370) y=(0xffff -y)*-1;

  return true;
                                                                     
  
}

//get the zoom setting

int cam::getZoom(int &z)
{
  unsigned char Inquirie[]={0x04,0x47};//ZoomPosInq
  
  int rxedDatalen;
  unsigned char Data[17];
  
  //Inquirie to the position of the zoom lens
  //give up if it fails
  if(!myViscaPTR->send_inquirie(1,Inquirie,rxedDatalen,Data)) return false; 
  if(rxedDatalen!=4) //must be 2 bytes
    {
      cerr <<"cam,getZoom, invalid reply size "<< rxedDatalen <<endl;
      return false; 
    }
  //the format of the data is.. 0z 0z 0z 0z

  //extract the z pos from the received data
  z = (0x0f&Data[3]) | ((0x0f&Data[2])<<4) | ((0x0f&Data[1])<<8) | ((0x0f&Data[0])<<12);

  return true;
                                                                     
  
}








//wait for the any moves and zooms to complete by repeatedly asking the cam for there
//positions and then compairing that possition to the last send pos.



int cam::waitComplete()
{
 int x,y,z;

 do
   {
     if(!(getPos(x,y) &getZoom(z))) return false; //fail if coms fault
   
     //add time out here!!
   }
 while ((x!=lastX)||(y!=lastY)||((-1>(z-lastZ))||((z-lastZ)>1))); 
 
 //I guess the people at sony wernt to fussed about z being bang on , it never quite 
 //gets there!! Hence the leway above.
   
 //move has compleated, and there where no communications faults.
  return true;
}
